// File for Belle2 slow control access with PCIe40 
#include "pcie40_b2slc.h"
#include <unistd.h>
#include <stdio.h>
#include "pcie40_ecs.h"

/* ---------------------------------------------------------------------- *\
   readfee8
   returns -1 in case of error
\* ---------------------------------------------------------------------- */
int pcie40_readfee8( int dev , int ch , int adr) {
  // PCIe40
  if ( ( adr <=0 ) || ( adr > 0x7F ) ) return -1 ;
  // Reset the FIFO (Emit and reception)
  unsigned ret = 0 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_RESET_ADD , 
		   1 << SLC_WFIFO_RESET_BIT ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_RESET_ADD , 0 ) ;
  if ( ret != 0 ) return -1 ;

  // Fill the FIFO with the requested information: write MSB first
  int data_word_1 = 0xFFFE ;
  int data_word_2 = 0x0001 ;
  int data_word_3 = 0x8000 | ( adr & 0x7F ) ;
  int data_word_4 = 0x000c ;
  int data_word_5 = 0xEEEE ;

  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_1 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_2 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_3 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_4 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_5 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;

  // start emit
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_START_ADD , 0 ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_START_ADD , 
  		   1 << ( SLC_WFIFO_EMIT_BIT + ch ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_START_ADD , 0 ) ;
  if ( ret != 0 ) return -1 ;

  // Wait for the result to come back (1 word in the fifo)
  int i ;
  for ( i=0 ; i<10 ; i++ ) {
    usleep( 10 ) ; //10 ms
    // TO DO: ret = ecs_read( dev , SLC_BAR , SLC_RFIFO_STATUS + ch * 0x40 ) ;
    // TO DO: if ( ret == 1 ) break;
  }

  // TO DO : if (i == 10) return -1;

  // Read the value 
  ret = ecs_read( dev , SLC_BAR , SLC_RFIFO_ADD + ch * 0x20 );
  
  // check address consistency
  if ( ( ( ret & 0x7F00 ) >> 8 ) != adr ) return -1 ;

  return ( ret & 0xFF ) ;
}

//==============================================================================
// Write 8b
//==============================================================================
int pcie40_writefee8( int dev , int ch , int adr , int val ) {
  // PCIe40
  if ( ( adr <=0 ) || ( adr > 0x7F ) ) return -1 ;
  // Reset the FIFO
  unsigned ret = 0 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_RESET_ADD , 
		   1 << SLC_WFIFO_RESET_BIT ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_RESET_ADD , 0 ) ;
  if ( ret != 0 ) return -1 ;

  // Fill the FIFO with the requested information: write MSB first
  int data_word_1 = 0xFFFF ;
  int data_word_2 = 0x0001 ;
  int data_word_3 = 0x8000 | ( adr & 0x7F ) ;
  int data_word_4 = 0x0000 | ( val & 0xFF ) ;
  int data_word_5 = 0xEEEE ;

  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_1 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_2 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_3 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_4 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_5 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;

  // start emit
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_START_ADD , 0 ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_START_ADD , 
  		   1 << ( SLC_WFIFO_EMIT_BIT + ch ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_START_ADD , 0 ) ;
  if ( ret != 0 ) return -1 ;

  return 0 ;
}

//==============================================================================
// Read 32b
//==============================================================================
unsigned long pcie40_readfee32( int dev , int ch , int adr ) {
  // PCIe40
  if ( ( adr <0 ) || ( adr > 0xFFFF ) ) return -1 ;
  // Reset the FIFO
  unsigned ret = 0 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_RESET_ADD , 
		   1 << SLC_WFIFO_RESET_BIT ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_RESET_ADD , 0 ) ;
  if ( ret != 0 ) return -1 ;

  // Fill the FIFO with the requested information: write MSB first
  int data_word_1 = 0xFFFB ;
  int data_word_2 = 0x7000 | ( ( adr & 0xFF00 ) >> 8 ) ;
  int data_word_3 = 0x7000 | ( adr & 0xFF ) ;
  int data_word_5 = 0xEEEE ;

  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_1 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_2 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_3 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_5 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;

  // start emit
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_START_ADD , 0 ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_START_ADD , 
  		   1 << ( SLC_WFIFO_EMIT_BIT + ch ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_START_ADD , 0 ) ;
  if ( ret != 0 ) return -1 ;

  // Wait for the result to come back
  int i ;
  for ( i=0 ; i<10 ; i++ ) {
    usleep( 100 ) ; //10 ms
    // TO DO ret = ecs_read( dev , SLC_BAR , SLC_RFIFO_STATUS + ch * 0x40 ) ;
    // TO DO if ( ret == 7 ) break;
  }

  // TO DO if (i == 10) return -1;

  // Read the value 
  int t_ret1 = ecs_read( dev , SLC_BAR , SLC_RFIFO_ADD + ch * 0x20 );
  int t_ret2 = ecs_read( dev , SLC_BAR , SLC_RFIFO_ADD + ch * 0x20 );
  int t_ret3 = ecs_read( dev , SLC_BAR , SLC_RFIFO_ADD + ch * 0x20 );
  int t_ret4 = ecs_read( dev , SLC_BAR , SLC_RFIFO_ADD + ch * 0x20 );  
  int t_ret5 = ecs_read( dev , SLC_BAR , SLC_RFIFO_ADD + ch * 0x20 );
  int t_ret6 = ecs_read( dev , SLC_BAR , SLC_RFIFO_ADD + ch * 0x20 );

  if ( ( t_ret1 & 0x0000FFFF ) != 0xFFFF ) {
    printf( "Bad header\n" ) ;
    return -1 ; 
  }
  if ( ( t_ret6 & 0x0000FFFF ) != 0xEEEE ) { 
    printf( "Bad trailer\n" ) ; 
    return -1 ;
  }

  int add1 = ( t_ret2 & 0xFF00 ) >> 8 ;
  int add2 = ( t_ret3 & 0xFF00 ) >> 8 ;
  int add3 = ( t_ret4 & 0xFF00 ) >> 8 ;
  int add4 = ( t_ret5 & 0xFF00 ) >> 8 ;
  
  if ( ( add1 & 0xF8 ) != 0x10 ) { 
    printf( "Bad address1 %X\n" , add1 ) ; 
    return -1 ;
  }
  if ( ( add2 & 0xF8 ) != 0x10 ) { 
    printf( "Bad address2 %X\n" , add2 ) ; 
    return -1 ;
  }
  if ( ( add3 & 0xF8 ) != 0x10 ) {
    printf( "Bad address3 %X\n" , add3 ) ; 
    return -1 ;
  }
  if ( ( add4 & 0xF8 ) != 0x10 ) {
    printf( "Bad address4 %X\n" , add4 ) ; 
    return -1 ;
  }

  add1 += 0x58 ;
  add2 += 0x58 ;
  add3 += 0x58 ;
  add4 += 0x58 ;
  
  if ( ( add2 != 0x6C ) || ( add1 != 0x6D ) || ( add4 != 0x6A ) || ( add3 != 0x6B ) )
    {
      printf( "Unexpected address %X %X %X %X\n" , add1 , add2 , add3 , add4 ) ;
      return -1 ;
    }

  int res1 = ( t_ret2 & 0xFF ) ;
  int res2 = ( t_ret3 & 0xFF ) ;
  int res3 = ( t_ret4 & 0xFF ) ;
  int res4 = ( t_ret5 & 0xFF ) ;

  return 0xFFFFFFFF & ( ( res1 << 24 ) | ( res2 << 16 ) | ( res3 << 8 ) | ( res4 ) ) ;
}

//==============================================================================
// Write 32b
//==============================================================================
int pcie40_writefee32( int dev , int ch , int adr , int val ) { 
  // PCIe40
  if ( ( adr <0 ) || ( adr > 0xFFFF ) ) return -1 ;
  // Reset the FIFO
  unsigned ret = 0 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_RESET_ADD , 
		   1 << ( SLC_WFIFO_RESET_BIT ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_RESET_ADD , 0 ) ;
  if ( ret != 0 ) return -1 ;

  int val1 = ( val & 0xFF ) ;
  int val2 = ( val & 0xFF00 ) >> 8 ;
  int val3 = ( val & 0xFF0000 ) >> 16 ;
  int val4 = ( val & 0xFF000000 ) >> 24 ;

  // Fill the FIFO with the requested information: write MSB first
  int data_word_1 = 0xFFFC ;
  int data_word_2 = 0x7000 | ( ( adr & 0xFF00 ) >> 8 ) ;
  int data_word_3 = 0x7000 | ( adr & 0xFF ) ;
  int data_word_4 = 0x7000 | val4 ;
  int data_word_5 = 0x7000 | val3 ;
  int data_word_6 = 0x7000 | val2 ;
  int data_word_7 = 0x7000 | val1 ;
  int data_word_8 = 0xEEEE ;


  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_1 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_2 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_3 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_4 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_5 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_6 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_7 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20 , 
		   (int) ( data_word_8 & 0xFFFFFFFF ) ) ;
  if ( ret != 0 ) return -1 ;

  // start emit
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_START_ADD , 0 ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_START_ADD , 
  		   1 << ( SLC_WFIFO_EMIT_BIT + ch ) ) ;
  if ( ret != 0 ) return -1 ;
  ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_START_ADD , 0 ) ;
  if ( ret != 0 ) return -1 ;

  return ret;
}

int pcie40_writestream( int dev , int ch , char * filename ) {
  return 0 ;
}
